/*-----------------------------------------------
 * xen/fbcrypt/pv/get_target_mfn.c
 *
 * XenBus から 任意のパスのmfnを取得・VMMに登録
 * ---------------------------------------------*/

#include <xen/lib.h>
#include <xen/types.h>
#include <xen/domain_page.h> 	
#include <xen/string.h>

#include <crypto/fbcrypt/pv/pv.h>	

/* XenBusリング視覚化 ON:1 OFF:0 */
#define DEBUG_XENSTORE_RING 0
#define DEBUG_XENSTORE_RING_INFO 0

/* XenStoreリングから取得する値のパス */
#define MFN_PATH_KEY 10
static char *target_key[MFN_PATH_KEY] = {
		"device/vfb/0/page-ref",
		"device/vkbd/0/page-ref",
		"device/vbd/51712/ring-ref",
		"device/vif/0/tx-ring-ref",
		"device/vif/0/rx-ring-ref",
};

/* Inter-domain shared memory communications. */
#define XENSTORE_RING_SIZE 1024
#define MASK_XENSTORE_IDX(idx) ((idx) & (XENSTORE_RING_SIZE-1))
typedef uint32_t XENSTORE_RING_IDX;

typedef struct xenstore_domain_interface {
    char req[XENSTORE_RING_SIZE]; /* Requests to xenstore daemon. */
    char rsp[XENSTORE_RING_SIZE]; /* Replies and async watch events. */
    XENSTORE_RING_IDX req_cons, req_prod;
    XENSTORE_RING_IDX rsp_cons, rsp_prod;
} xs_interface;


/* fbcrypt/domain_info.c */
int XS_header_on(dominfo_p p, XS_DATA cmd);
int XS_get_clue(dominfo_p p, XS_DATA cmd, xs_mfn_flag *clue);
int XS_plus_cntWrite(dominfo_p p);
int XS_plus_cntAll(dominfo_p p);
int XS_save_memData(dominfo_p p, XS_DATA cmd, 
XENSTORE_RING_IDX cons, XENSTORE_RING_IDX prod, unsigned avail);


static int check_indexes(XENSTORE_RING_IDX cons, XENSTORE_RING_IDX prod)
{
	return ((prod - cons) <= XENSTORE_RING_SIZE);
}


/* DomainUの書き込みリング(req)の読み込み開始位置を返す関数 */
/* from tools/xenstore/xenstored_domain.c/readchn */
static const void *get_input_chunk(XENSTORE_RING_IDX cons,
				   XENSTORE_RING_IDX prod,
				   const char *buf, uint32_t *len)
{
	*len = XENSTORE_RING_SIZE - MASK_XENSTORE_IDX(cons);
	if ((prod - cons) < *len)
		*len = prod - cons;
	return buf + MASK_XENSTORE_IDX(cons);
}


/* XenBus RING Visuable */
#if DEBUG_XENSTORE_RING
static int xb_ring_debug(char data[])
{
	char body[XENSTORE_RING_SIZE] = {'Z'};
	int i = 0;

	memcpy(body,data,XENSTORE_RING_SIZE);

	/* NULL文字を'Z'に置換 */
	for ( i = 0; i < XENSTORE_RING_SIZE; i++ ) {
		if ( body[i] == '\0' ) 
			body[i] = 'Z';
	}

	body[XENSTORE_RING_SIZE-1] = '\0';
	printk("body: %s",body);

	return 0;
}
#endif


/* K&R Version 2 */
static int my_atoi(const char *str)
{
	int cnt;
	int num = 0;
	for ( cnt = 0; (str[cnt] >= '0') && (str[cnt] <= '9'); cnt++ )
		num = 10 * num + (str[cnt] - '0');
	return num;
}

/* keyと対応する値を取得してVMMに登録 */
/* [方針]
 * keyの値とsrcが一致していたらheader_flagをON
 * dstにsrcをavailサイズだけコピーする
 * dstにはtarget_mfnが入る
 * dstは文字なので数値に変換する
 */
static int xs_search_mfn(int domid, dominfo_p p, XS_DATA cmd, char *key) 
{
	xs_mfn_flag clue = {0};			// domain_info から情報をもらう
	unsigned long target_mfn = 0; 	// cmd で指定したmfnが入る

	xs_interface *intf = NULL;		// DomainUのXenStoreリングのMFN
	const char *src = NULL;	  	  	// reqリングの読み込み開始位置ポインタ
	char dst[100] = {'\0'};	  	  	// srcのコピー先
	unsigned int avail = 0;	  	  	// 読み取り可能バイト数
	unsigned int subtract = 0;		// consとprodの差
    XENSTORE_RING_IDX cons;	  	  	// consumer: Domain0@xenstored
    XENSTORE_RING_IDX prod;   	  	// producer: DomainU@kernel

	/* xenbusリングを取得 */
	intf = Domain_mfn2map_page(domid,VMM_STORE);
	cons = intf->req_cons;
	prod = intf->req_prod;
	subtract = prod-cons;

	// subtractが0 -> 書き込みが行われていない. 
	// もしくは xb_write以外のイベント(xb_readなど)が発行されている
	// ---> 処理中断
	if ( !check_indexes(cons, prod) || subtract == 0 ) {
		unmap_domain_page(intf);
		return 1;
	}

	XS_plus_cntWrite(p);

	
	/* domain_info.c にあるXenStoreリング探索に必要なデータを取得 */
	XS_get_clue(p,cmd,&clue);
	
	#if DEBUG_XENSTORE_RING_INFO
	printk("\n################### STRAT ###########################\n");
	#endif

	#if DEBUG_XENSTORE_RING  
	printk("\nXB_RING_Visualize =======================\n");
	xb_ring_debug(intf->req);
	printk("\nEND XB_RING_Visualize ===================\n\n");
	#endif 

	/* ================================================================ header_flag OFF */
	if ( clue.header_flag == 0 ) {
		/*-----------------------------------------------------------TYPE B */
		// ヘッダ未読状態
		if ( cons == clue.mem_cons ) {
			// ヘッダをスルーしてPATHを読む
			// ヘッダをとばした読み込み開始地点をsrcに
			src = get_input_chunk( cons+(clue.mem_avail),prod,intf->req,&avail );
			// パスのみをdstにコピー
			memcpy( dst,src,subtract-(clue.mem_avail) );
		} else
		/*-----------------------------------------------------------TYPE A & C */
		// ヘッダ取得済み
		if ( cons != clue.mem_cons ) {
			src = get_input_chunk(cons,prod,intf->req,&avail);
			memcpy(dst,src,avail);
		}
		/*----------------------------------------------------------- 共通処理 */
		// dstとkey(PATH)が一致したらheader_flagをON
		// header_flag がONの場合、headerは既にリングに書き込み済みということ
		if ( !memcmp(dst,key,strlen(key)) ) {
			#if DEBUG_XENSTORE_RING_INFO
			printk("get_target_mfn.c: SRC & TARGET_KEY MATCH\n");
			printk("Domain:%d KEY:%s\n",domid,key);
			#endif
			XS_header_on(p,cmd); /* ヘッダ読み込み完了フラグON */
		}
		/* cons,prod,avail の値を各ドメインごとにcmdで識別して保存 */
		XS_save_memData(p,cmd,cons,prod,avail);	
	} else
	
	/* ================================================================ header_flag ON */
	if ( clue.header_flag == 1 ) {
		/*-----------------------------------------------------------TYPE C */
		// PATH未読
		if ( cons == clue.mem_cons ) {
			// PATHをスルーしてValueを読む
			// PATHをとばした読み込み開始地点をsrcに
			src = get_input_chunk( cons+(clue.mem_avail),prod,intf->req,&avail );
			// Valueのみをdstにコピー
			memcpy( dst,src,subtract-(clue.mem_avail) );
		} else 
		// PATHがmatch済みなら
		if ( cons != clue.mem_cons ) {
			src = get_input_chunk(cons,prod,intf->req,&avail);
			memcpy(dst,src,avail);
			/*-----------------------------------------------------------TYPE B */
			// srcとPATHが一致したなら
			if ( !memcmp(dst,key,strlen(key)) ) {
				// dstのPATHのデータサイズ分とばしてValueをよむ
				src = get_input_chunk(cons+strlen(key),prod,intf->req,&avail);
				memcpy( dst,src,subtract-strlen(key) );
			}
		 	else {
				/*-----------------------------------------------------------TYPE D */
				// VMMがキャッチできない書き込みがあったならすでにパスが書き込まれている
				// 値と次のヘッダが既に書き込まれている場合
				// 前回書き込んだサイズより読み取ったサイズの方が大きいなら...
			 	if ( cons > clue.mem_prod ) { 
					src = get_input_chunk( clue.mem_prod,prod,intf->req,&avail );
					memcpy( dst,src,cons-(clue.mem_prod) );
				} else {
				/*-----------------------------------------------------------TYPE A */
					// srcがPATH->EV->Value という単純なパターン
					src = get_input_chunk(cons,prod,intf->req,&avail);
					memcpy(dst,src,avail);
				}
			}
		}
			/*------------------------------------------------ header_flag ==1 共通 */
			target_mfn = my_atoi(dst);	// 文字列を数値に変換

			if ( target_mfn != 0 ) {
				Domain_regist_info(domid,target_mfn,cmd); // MFNをVMMに登録
				Domain_ref_list(domid); 
			}
	}

	/*----------------------------------------------------------- 共通処理 */
	#if DEBUG_XENSTORE_RING_INFO
	printk("XB_RING INFORMATION =====================\n");
	printk("intf->req_cons: %d\n",cons);
	printk("intf->req_prod: %d\n",prod);
	printk("subtract: %d\n",subtract);
	printk("avail: %d\n",avail);
	printk("src: %s\n",src);
	printk("dst: %s\n",dst);
	printk("#################### END ############################\n");
	#endif 

	unmap_domain_page(intf);
	return 0;
}


/* Domain0からのイベントでは実行されない(common/event_channel.c) */
int get_target_mfn(int domid)
{
	dominfo_p p = NULL;
	unsigned long store_mfn	 = Domain_ref_mfn(domid,VMM_STORE);
	unsigned long vfb_mfn  	 = Domain_ref_mfn(domid,VMM_VFB);
	unsigned long vkbd_mfn 	 = Domain_ref_mfn(domid,VMM_VKBD);
	//unsigned long vbd_mfn  = Domain_ref_mfn(domid,VMM_VBD);
	unsigned long vif_tx_mfn = Domain_ref_mfn(domid,VMM_VIF_TX);
	unsigned long vif_rx_mfn = Domain_ref_mfn(domid,VMM_VIF_RX);

	/* store_mfnが初期値のまま-> リングのチェックをせず終了 */
	if ( store_mfn == 0 ) return 1;

	/* 何度もリストから検索するのは効率悪いので
	 * 対象のドメインのポインタを取得して使いまわす */
	p = Domain_get_pointer(domid);
	if ( p == NULL ) {
		printk("ERROR: get_target_mfn p==NULL\n");
		return 1;
	}

	//XS_plus_cntAll(p);

	/* cmdで指定されたmfnをサーチしてVMMに登録まで */
	/* xs_search_mfn内でpの内部は隠蔽されてることに注意 */
	if ( vfb_mfn  == 0 ) xs_search_mfn(domid,p,VMM_VFB,target_key[0]);
	if ( vkbd_mfn == 0 ) xs_search_mfn(domid,p,VMM_VKBD,target_key[1]);
	//if ( vbd_mfn  == 0 ) xs_search_mfn(domid,p,VMM_VBD,target_key[2]);
	if ( vif_tx_mfn  == 0 ) xs_search_mfn(domid,p,VMM_VIF_TX,target_key[3]);
	if ( vif_rx_mfn  == 0 ) xs_search_mfn(domid,p,VMM_VIF_RX,target_key[4]);

	return 0;
}


// 本当は以下のようにxs_search_mfn はmfnを返すようにしたいが
// うまく動作しない。vkbd_mfn がどこかで書き換えられて1になってしまう
#if 0
	/* cmdで指定されたmfnをサーチしてVMMに登録まで */
	if ( vkbd_mfn == 0 ) { /* VKBD */
		vkbd_mfn = xs_search_mfn(domid,VMM_VKBD,target_key[0]);
		printk("NOTIFY!!!! vkbd_mfn:%ld\n",vkbd_mfn);
		if ( vkbd_mfn != 0 ) {
			Domain_regist_info(domid,vkbd_mfn,VMM_VKBD); 	// MFNをVMMに登録
			Domain_ref_list(domid); 
		}
	}

	if ( vfb_mfn  == 0 ) { /* VFB */
		vfb_mfn = xs_search_mfn(domid,VMM_VFB,target_key[1]);
		if ( vfb_mfn != 0 ) {
			Domain_regist_info(domid,vfb_mfn,VMM_VFB);
			Domain_ref_list(domid); 
		}
	}

	if ( vif_tx_mfn  == 0 ) { /* VIF_TX */
		vif_tx_mfn = xs_search_mfn(domid,VMM_VIF_TX,target_key[2]);
		if ( vif_tx_mfn != 0 ) {
			Domain_regist_info(domid,vif_tx_mfn,VMM_VIF_TX);
			Domain_ref_list(domid); 
		}
	}

	if ( vif_rx_mfn  == 0 ) { /* VIF_RX */
		vif_rx_mfn = xs_search_mfn(domid,VMM_VIF_RX,target_key[3]);
		if ( vif_rx_mfn != 0 ) {
			Domain_regist_info(domid,vif_rx_mfn,VMM_VIF_RX);
			Domain_ref_list(domid); 
		}
	}
#endif


